static const char *RcsId = "$Header: /home/cvsadm/cvsroot/fermi/servers/a2605/src/A2605.cpp,v 1.24 2019-04-15 06:28:40 giulio Exp $";
//+=============================================================================
//
// file :         A2605.cpp
//
// description :  C++ source for the A2605 and its commands. 
//                The class is derived from Device. It represents the
//                CORBA servant object which will be accessed from the
//                network. All commands which can be executed on the
//                A2605 are implemented in this file.
//
// project :      TANGO Device Server
//
// $Author: giulio $
//
// $Revision: 1.24 $
//
// $Log: A2605.cpp,v $
// Revision 1.24  2019-04-15 06:28:40  giulio
// Moved to tango9
//
// Revision 1.23  2018-05-08 06:51:03  giulio
// Do not go in read/set error when generating a waveform
//
// Revision 1.22  2018-05-03 12:07:35  giulio
// Added running status to allow current DC setting during waveform generation
//
// Revision 1.21  2013-08-23 10:24:44  giulio
// Moved to Device4Impl
//
// Revision 1.20  2013-08-22 12:12:18  giulio
// Moved to Tango8
//
// Revision 1.19  2013-04-16 10:21:54  giulio
// Fixed get_current_set bug
//
// Revision 1.18  2013-04-16 09:44:23  giulio
// Update rt-buffer api's, removed tango rpath, added DeviceState
//
// Revision 1.17  2012-12-27 10:45:58  giulio
// Added tcp communication support
//
// Revision 1.16  2012-04-19 09:38:45  giulio
// Logging command disabled when realtime is disabled
//
// Revision 1.15  2012-04-19 09:28:22  giulio
// Added command logging buffer
//
// Revision 1.14  2012-02-02 07:48:52  giulio
// Do not set current limits when not in realtime control
//
// Revision 1.13  2012-01-02 12:55:29  giulio
// .Changed logic in local+rnm current control, add degauss, updated cycling procedure
//
// Revision 1.12  2011/08/23 14:00:30  giulio
// .Diff error on only in local mode
//
// Revision 1.11  2011/01/21 13:15:38  giulio
// .Fixed save property issue
//
// Revision 1.10  2011/01/17 13:30:20  giulio
// .Set and save ps mac address
//
// Revision 1.9  2011/01/17 11:12:44  giulio
// .Off managed by thread
//
// Revision 1.8  2010/10/12 06:36:58  giulio
// .Fixed read current write part in slow mode
//
// Revision 1.7  2010/09/27 12:01:42  giulio
// .Check set/read difference.Fault state when difference above a threshold for a certain amount of time
//
// Revision 1.6  2010/09/20 06:55:57  giulio
// .Added support for non realtime communication protocol
//
// Revision 1.5  2010/09/07 15:45:05  giulio
// .Fixed bug in overcurrent notification fault
//
// Revision 1.4  2010/08/31 11:41:15  giulio
// Added support up to 64 channels
//
// Revision 1.3  2010/08/26 13:11:24  giulio
// Fixed a bug in get_wave api
//
//
// copyleft :     European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                FRANCE
//
//-=============================================================================
//
//  		This file is generated by POGO
//	(Program Obviously used to Generate tango Object)
//
//         (c) - Software Engineering Group - ESRF
//=============================================================================



//===================================================================
//
//	The following table gives the correspondence
//	between commands and method name.
//
//  Command name     |  Method name
//	----------------------------------------
//  State            |  dev_state()
//  Status           |  dev_status()
//  On               |  on()
//  Off              |  off()
//  Reset            |  reset()
//  GetCurrent       |  get_current()
//  GetVoltage       |  get_voltage()
//  GetVdcLink       |  get_vdc_link()
//  WaveStartSw      |  wave_start_sw()
//  StartCycling     |  start_cycling()
//  Abort            |  abort()
//  GetCurrentSet    |  get_current_set()
//  SaveConfig       |  save_config()
//  StartDegaussing  |  start_degaussing()
//  GetCmdLog        |  get_cmd_log()
//
//===================================================================


#include <tango.h>
#include <A2605.h>
#include <A2605Class.h>
#include "cyclethread.h"
#include "offthread.h"

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <vector>

#include <check_get_rt_buffer.h>
#include <devicestate.h>

namespace A2605_ns
{

//+----------------------------------------------------------------------------
//
// method : 		A2605::A2605(string &s)
// 
// description : 	constructor for simulated A2605
//
// in : - cl : Pointer to the DeviceClass object
//      - s : Device name 
//
//-----------------------------------------------------------------------------
A2605::A2605(Tango::DeviceClass *cl,string &s)
:TANGO_BASE_CLASS(cl,s.c_str())
{
	init_device();
}

A2605::A2605(Tango::DeviceClass *cl,const char *s)
:TANGO_BASE_CLASS(cl,s)
{
	init_device();
}

A2605::A2605(Tango::DeviceClass *cl,const char *s,const char *d)
:TANGO_BASE_CLASS(cl,s,d)
{
	init_device();
}
//+----------------------------------------------------------------------------
//
// method : 		A2605::delete_device()
// 
// description : 	will be called at device destruction or at init command.
//
//-----------------------------------------------------------------------------
void A2605::delete_device()
{
	//	Delete device's allocated object

	if (realtimeControl) {
		a2605_set_rtmode(psIndex, 0, a2605); /* disable realtime communication */
	
		if (a2605_deinit(a2605) < 0)
			printf("Failed to remove realtime shared memory\n");
	
		delete_buffers();		
	}
	else {
		delete_udp_socket();
	}

	delete in_mutex;
	
	delete attr_CurrentSet_read;
	delete attr_Current_read;
	delete attr_Voltage_read;
	delete attr_VdcLink_read;
	delete attr_CurrentFb_read;
	delete attr_CommFrequency_read;
	
  if (*attr_PsMacAddress_read) 
    delete *attr_PsMacAddress_read;  
  if (attr_PsMacAddress_read) {
	  delete[] attr_PsMacAddress_read;
    attr_PsMacAddress_read = 0;
  }
			
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::init_device()
// 
// description : 	will be called at device initialization.
//
//-----------------------------------------------------------------------------
void A2605::init_device()
{
	INFO_STREAM << "A2605::A2605() create device " << device_name << endl;

	// Initialise variables to default values
	//--------------------------------------------
	get_device_property();

	Tango::Database *db = new Tango::Database();
	try {
		Tango::DbData db_data;
		db_data.push_back((Tango::DbDatum("Host")));
		db_data.push_back((Tango::DbDatum("Port")));
		db->get_property("Database",db_data);
		db_data[0] >> host_rw;
		db_data[1] >> port_rw;
	}
	catch(Tango::DevFailed &e) {
	      cout << __FUNCTION__ << " Error reading Database property='" << e.errors[0].desc << "'";
	}
	delete db;

	/* if rtmodule not installed force to conventional comm. mode */
	a2605 = 0;
	polling_config = false;
	
	in_mutex = new omni_mutex::omni_mutex();

	if ((a2605 = a2605_init()) == NULL) {
		realtimeControl = false;
	}

	if (realtimeControl) {

		if (init_buffers() == false) {
			printf("Failed to initialize realtime buffer, exit\n");
			exit(0);
		}

		a2605 = 0;
		if ((a2605 = a2605_init()) == NULL) {
			printf("Failed to initialize realtime shared memory, exit\n");
			exit(0);
		}

		/* set bpm name */
		vector<string> results; 
		string_explode(this->get_name(), "/", &results);
		a2605_set_ps_label(psIndex, (char *)results[2].c_str(), rnmIndex, a2605);

		configure_network();

		/* enable realtime communication; default is 2A */
		a2605_set_min_delta_cur(psIndex, (int)(minDeltaCurRamp*1000000), a2605);

		/* enable realtime communication */
		a2605_set_rtmode(psIndex, 1, a2605); 
		/* enable slew rate control */
		if (slewRateLimit > 0)
			a2605_set_slewrate_limit(psIndex, 1, a2605); /* enables */
		else
			a2605_set_slewrate_limit(psIndex, 0, a2605); /* disable */ 
	}
	else {
		char_to_hex['k']=0; /* dummy value for exception detection */
		char_to_hex['0']=0x1;char_to_hex['1']=0x2;char_to_hex['2']=0x3;char_to_hex['3']=0x4;
		char_to_hex['4']=0x5;char_to_hex['5']=0x6;char_to_hex['6']=0x7;char_to_hex['7']=0x8;
		char_to_hex['8']=0x9;char_to_hex['9']=0xA;char_to_hex['A']=0xB;char_to_hex['B']=0xC;
		char_to_hex['C']=0xD;char_to_hex['D']=0xE;char_to_hex['F']=0xF;char_to_hex['F']=0x10;
		gettimeofday(&last_sock_time,NULL);
		if (create_udp_socket() == -1)	
			last_sock_time.tv_sec -= 100;
	}

	for (int i = 0; i < A2605_NUM_STAT; i++)
		PsStat[i] = false;
		
	cycle_abort = false;
	cycle_run = false;
	off_abort = false;
	off_run = false;
	
	attr_CurrentSet_read = new Tango::DevDouble();
	attr_Current_read = new Tango::DevDouble();
	attr_Voltage_read = new Tango::DevDouble();
	attr_VdcLink_read = new Tango::DevDouble();
	attr_CurrentFb_read = new Tango::DevDouble();
	attr_CommFrequency_read = new Tango::DevDouble();	

	attr_PsMacAddress_read = new char*;
	*attr_PsMacAddress_read = new char[256];
  attr_PsMacAddress_write = 0;	

	/* used to force current attribute set value */
	try {
		Tango::MultiAttribute *attr_list = get_device_attr();    
		attr_set_current = 0;
		attr_set_current = &(attr_list->get_w_attr_by_name("Current"));
		if (!realtimeControl) {
			double curr;
			get_udp_current(&curr);
			attr_set_current->set_write_value(curr);
			attr_Current_write = curr;
		}
	}
	catch(...) {
		INFO_STREAM << "A2605::init_device(): ERROR GETTING ATTRIBUTE" << endl;
	} 		

	/* set values as CurrentControl attribute property */	
	try {
  	Tango::DbDatum db_curctrl("CurrentControl"); 
  	db_curctrl << (short)1;
		Tango::DbDatum db_values("values");
		Tango::DbData db_data_set;
  	vector<string> str_values;
  	str_values.push_back("0");
  	str_values.push_back("1");
  	str_values.push_back("2");		
  	db_values << str_values;
		db_data_set.push_back(db_curctrl);
  	db_data_set.push_back(db_values);
		get_db_device()->put_attribute_property(db_data_set);
	}
	catch(...) {
		INFO_STREAM << "A2605::init_device(): ERROR PUTTING ATTRIBUTE PROPERTY" << endl;
	};

	if (realtimeControl) {
		/* enable realtime communication */
		a2605_set_current_control(psIndex, (long)currentControl, a2605); 	
	}
	
	gettimeofday(&last_data_time,NULL);
	last_data_time.tv_sec -= 100;

	udp_cycle_flag = false;
	udp_off_flag = false;	
	
	gettimeofday(&curr_read_diff_time,NULL);
	diff_err = 0;

	/* get min/max current */
	Tango::DevVarStringArray attrnames2(1);
  attrnames2.length(1); 
	attrnames2[0] = CORBA::string_dup("Current"); 
	Tango::AttributeConfigList_var attr_config_list; 
  attr_config_list = get_attribute_config(attrnames2);
	for(unsigned int i = 0; i < attr_config_list->length(); i++)  {
		std::stringstream s_min, s_max;
		s_min << attr_config_list[i].min_value;
		s_max << attr_config_list[i].max_value;
		double cur_min, cur_max;
		s_min >> cur_min;
		s_max >> cur_max;
		cur_min *= 1000000;
		cur_max *= 1000000;
		if (realtimeControl == true)
			a2605_set_minmax_current(psIndex, (long) cur_min, (long) cur_max, a2605);
	}

	degauss_flag = false;

}


//+----------------------------------------------------------------------------
//
// method : 		A2605::get_device_property()
// 
// description : 	Read the device properties from database.
//
//-----------------------------------------------------------------------------
void A2605::get_device_property()
{
	//	Initialize your default values here (if not done with  POGO).
	//------------------------------------------------------------------

	//	Read device properties from database.(Automatic code generation)
	//------------------------------------------------------------------
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("PsIp"));
	dev_prop.push_back(Tango::DbDatum("PsMacAddress"));
	dev_prop.push_back(Tango::DbDatum("PsUdpPort"));
	dev_prop.push_back(Tango::DbDatum("CpuUdpPort"));
	dev_prop.push_back(Tango::DbDatum("PsModel"));
	dev_prop.push_back(Tango::DbDatum("PsIndex"));
	dev_prop.push_back(Tango::DbDatum("RnmIndex"));
	dev_prop.push_back(Tango::DbDatum("CpuIp"));
	dev_prop.push_back(Tango::DbDatum("SlewRateLimit"));
	dev_prop.push_back(Tango::DbDatum("CurrentControl"));
	dev_prop.push_back(Tango::DbDatum("RealtimeControl"));
	dev_prop.push_back(Tango::DbDatum("DiffTimeErr"));
	dev_prop.push_back(Tango::DbDatum("DiffCurrErr"));
	dev_prop.push_back(Tango::DbDatum("MinDeltaCurRamp"));

	//	Call database and extract values
	//--------------------------------------------
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->get_property(dev_prop);
	Tango::DbDatum	def_prop, cl_prop;
	A2605Class	*ds_class =
		(static_cast<A2605Class *>(get_device_class()));
	int	i = -1;

	//	Try to initialize PsIp from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  psIp;
	else {
		//	Try to initialize PsIp from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  psIp;
	}
	//	And try to extract PsIp value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  psIp;

	//	Try to initialize PsMacAddress from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  psMacAddress;
	else {
		//	Try to initialize PsMacAddress from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  psMacAddress;
	}
	//	And try to extract PsMacAddress value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  psMacAddress;

	//	Try to initialize PsUdpPort from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  psUdpPort;
	else {
		//	Try to initialize PsUdpPort from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  psUdpPort;
	}
	//	And try to extract PsUdpPort value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  psUdpPort;

	//	Try to initialize CpuUdpPort from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cpuUdpPort;
	else {
		//	Try to initialize CpuUdpPort from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cpuUdpPort;
	}
	//	And try to extract CpuUdpPort value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cpuUdpPort;

	//	Try to initialize PsModel from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  psModel;
	else {
		//	Try to initialize PsModel from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  psModel;
	}
	//	And try to extract PsModel value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  psModel;

	//	Try to initialize PsIndex from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  psIndex;
	else {
		//	Try to initialize PsIndex from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  psIndex;
	}
	//	And try to extract PsIndex value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  psIndex;

	//	Try to initialize RnmIndex from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  rnmIndex;
	else {
		//	Try to initialize RnmIndex from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  rnmIndex;
	}
	//	And try to extract RnmIndex value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  rnmIndex;

	//	Try to initialize CpuIp from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  cpuIp;
	else {
		//	Try to initialize CpuIp from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  cpuIp;
	}
	//	And try to extract CpuIp value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  cpuIp;

	//	Try to initialize SlewRateLimit from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  slewRateLimit;
	else {
		//	Try to initialize SlewRateLimit from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  slewRateLimit;
	}
	//	And try to extract SlewRateLimit value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  slewRateLimit;

	//	Try to initialize CurrentControl from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  currentControl;
	else {
		//	Try to initialize CurrentControl from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  currentControl;
	}
	//	And try to extract CurrentControl value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  currentControl;

	//	Try to initialize RealtimeControl from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  realtimeControl;
	else {
		//	Try to initialize RealtimeControl from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  realtimeControl;
	}
	//	And try to extract RealtimeControl value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  realtimeControl;

	//	Try to initialize DiffTimeErr from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  diffTimeErr;
	else {
		//	Try to initialize DiffTimeErr from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  diffTimeErr;
	}
	//	And try to extract DiffTimeErr value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  diffTimeErr;

	//	Try to initialize DiffCurrErr from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  diffCurrErr;
	else {
		//	Try to initialize DiffCurrErr from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  diffCurrErr;
	}
	//	And try to extract DiffCurrErr value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  diffCurrErr;

	//	Try to initialize MinDeltaCurRamp from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  minDeltaCurRamp;
	else {
		//	Try to initialize MinDeltaCurRamp from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  minDeltaCurRamp;
	}
	//	And try to extract MinDeltaCurRamp value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  minDeltaCurRamp;



	//	End of Automatic code generation
	//------------------------------------------------------------------

}
//+----------------------------------------------------------------------------
//
// method : 		A2605::always_executed_hook()
// 
// description : 	method always executed before any command is executed
//
//-----------------------------------------------------------------------------
void A2605::always_executed_hook()
{
	
	set_DevStat();
	
	
}
//+----------------------------------------------------------------------------
//
// method : 		A2605::read_attr_hardware
// 
// description : 	Hardware acquisition for attributes.
//
//-----------------------------------------------------------------------------
void A2605::read_attr_hardware(vector<long> &attr_list)
{
	DEBUG_STREAM << "A2605::read_attr_hardware(vector<long> &attr_list) entering... "<< endl;
	//	Add your own code here
}
//+----------------------------------------------------------------------------
//
// method : 		A2605::read_DeviceState
// 
// description : 	Extract real attribute values for DeviceState acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_DeviceState(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_DeviceState(Tango::Attribute &attr) entering... "<< endl;

	attr.set_value(&attr_DeviceState_write);

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_DeviceState
// 
// description : 	Write DeviceState attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_DeviceState(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_DeviceState(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_DeviceState_write);

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_PsMacAddress
// 
// description : 	Extract real attribute values for PsMacAddress acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_PsMacAddress(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_PsMacAddress(Tango::Attribute &attr) entering... "<< endl;
	
	strncpy(*attr_PsMacAddress_read, psMacAddress.c_str(), strlen(psMacAddress.c_str()));
	attr.set_value(attr_PsMacAddress_read);
		
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_PsMacAddress
// 
// description : 	Write PsMacAddress attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_PsMacAddress(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_PsMacAddress(Tango::WAttribute &attr) entering... "<< endl;

	attr.get_write_value(attr_PsMacAddress_write);
	string mac_address(attr_PsMacAddress_write);	

	int count_colon = 0;
	for (unsigned int i = 0; i < mac_address.length(); i++) {	
		if (mac_address.at(i) == ':')
			count_colon++;
	}
	if ((mac_address.length() != 17) || (count_colon != 5)) {
		TangoSys_MemStream out_stream;
		out_stream << "Mac address invalid format: " << mac_address << ends;
		Tango::Except::throw_exception (
			out_stream.str(),
			(const char *)"Failed to set ps mac address",
			(const char *)"A2605::write_PsMacAddress()");
	}
	
	psMacAddress.resize(0);
	psMacAddress = mac_address.c_str();
	
	configure_network();

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_CommFrequency
// 
// description : 	Extract real attribute values for CommFrequency acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_CommFrequency(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_CommFrequency(Tango::Attribute &attr) entering... "<< endl;
	
		if (realtimeControl) {
		*attr_CommFrequency_read = 1000000 / (double) a2605_get_comm_period(psIndex, a2605);
	}
	else {
		*attr_CommFrequency_read = 0;
	}	
	attr.set_value(attr_CommFrequency_read);	

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_CurrentSet
// 
// description : 	Extract real attribute values for CurrentSet acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_CurrentSet(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_CurrentSet(Tango::Attribute &attr) entering... "<< endl;
	
	if (realtimeControl) {
		long data;
		unsigned long bunchnumber;
		struct timeval buf_time;
	
		check_comm();
		read_last(&data, &buf_time, &bunchnumber, buf[A2605_CURRSET_IDX]);
		*attr_CurrentSet_read = ((double)data)/1000000;
	}
	else {
		/*get_udp_current(attr_CurrentSet_read);*/ /* no hw readback from ps, replace with analog one */
		*attr_CurrentSet_read = attr_Current_write;
		
	}
	attr.set_value(attr_CurrentSet_read);			
	
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_CurrentSet
// 
// description : 	Write CurrentSet attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_CurrentSet(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_CurrentSet(Tango::WAttribute &attr) entering... "<< endl;
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_BunchNumber
// 
// description : 	Extract real attribute values for BunchNumber acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_BunchNumber(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_BunchNumber(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		attr_l_value = a2605_get_bunchnumber(psIndex, a2605);
	}
	else {
		attr_l_value = 0;
	}
	attr.set_value(&attr_l_value);	

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_CurrentFb
// 
// description : 	Extract real attribute values for CurrentFb acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_CurrentFb(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_CurrentFb(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		*attr_CurrentFb_read = ((double)a2605->ps[psIndex].iset_rnm_wr)/1000000;
	}
	else {
		*attr_CurrentFb_read = 0;
	}
	attr.set_value(attr_CurrentFb_read);		

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveAbsCurMode
// 
// description : 	Extract real attribute values for WaveAbsCurMode acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveAbsCurMode(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveAbsCurMode(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		char value = a2605_get_wave_abs_cur_mode(psIndex, a2605);
		attr_b_value = (bool) value;
		attr.set_value(&attr_b_value);		
	}	

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveAbsCurMode
// 
// description : 	Write WaveAbsCurMode attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveAbsCurMode(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveAbsCurMode(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveAbsCurMode(): " << attr_WaveAbsCurMode_write << endl;

	if (realtimeControl) {
		attr.get_write_value(attr_WaveAbsCurMode_write);	
		char value;

		if (attr_WaveAbsCurMode_write == true)
			value = 1;
		else
			value = 0;
		a2605_set_wave_abs_cur_mode(psIndex, value, a2605);		
	
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveBunchNumEnd
// 
// description : 	Extract real attribute values for WaveBunchNumEnd acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveBunchNumEnd(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveBunchNumEnd(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		attr_l_value = a2605_get_bunch_number_end(psIndex, a2605);
		attr.set_value(&attr_l_value);		
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveBunchNumStart
// 
// description : 	Extract real attribute values for WaveBunchNumStart acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveBunchNumStart(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveBunchNumStart(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		attr_l_value = a2605_get_bunch_number_start(psIndex, a2605);
		attr.set_value(&attr_l_value);	
	}	

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveBunchNumStart
// 
// description : 	Write WaveBunchNumStart attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveBunchNumStart(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveBunchNumStart(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveBunchNumStart(): " << attr_WaveBunchNumStart_write << endl;

	if (realtimeControl) {
		attr.get_write_value(attr_WaveBunchNumStart_write);		
	
		if (a2605_set_bunch_number_start(psIndex, (long) attr_WaveBunchNumStart_write, a2605) < 0) {
			Tango::Except::throw_exception (
				(const char *)"Bunch number not valid",
				(const char *)"Setting bunch number start failed",
				(const char *)"A2605::write_WaveBunchNumStart()");		
		}
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveCycles
// 
// description : 	Extract real attribute values for WaveCycles acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveCycles(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveCycles(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		attr_l_value = a2605_get_num_waves(psIndex, a2605);
		attr.set_value(&attr_l_value);
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveCycles
// 
// description : 	Write WaveCycles attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveCycles(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveCycles(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveCycles(): " << attr_WaveCycles_write << endl;
	
	if (realtimeControl) {
		attr.get_write_value(attr_WaveCycles_write);	
		a2605_set_num_waves(psIndex, (long)attr_WaveCycles_write, a2605);			
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveLoopMode
// 
// description : 	Extract real attribute values for WaveLoopMode acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveLoopMode(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveLoopMode(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		char value = a2605_get_wave_loop_mode(psIndex, a2605);
		attr_b_value = (bool) value;
		attr.set_value(&attr_b_value);		
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveLoopMode
// 
// description : 	Write WaveLoopMode attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveLoopMode(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveLoopMode(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveLoopMode(): " << attr_WaveLoopMode_write << endl;

	if (realtimeControl) {
		attr.get_write_value(attr_WaveLoopMode_write);	
		char value;
		if (attr_WaveLoopMode_write == true)
			value = 1;
		else
			value = 0;
		a2605_set_wave_loop_mode(psIndex, value, a2605);		
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveStepFactor
// 
// description : 	Extract real attribute values for WaveStepFactor acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveStepFactor(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveStepFactor(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		attr_l_value = a2605_get_wave_step_factor(psIndex, a2605);
		attr.set_value(&attr_l_value);
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveStepFactor
// 
// description : 	Write WaveStepFactor attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveStepFactor(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveStepFactor(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveStepFactor(): " << attr_WaveStepFactor_write << endl;

	if (realtimeControl) {
		attr.get_write_value(attr_WaveStepFactor_write);	
		a2605_set_wave_step_factor(psIndex, (long)attr_WaveStepFactor_write, a2605);			
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveTrigBunchNum
// 
// description : 	Extract real attribute values for WaveTrigBunchNum acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveTrigBunchNum(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveTrigBunchNum(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		char value = a2605_get_wave_trig_bnc(psIndex, a2605);
		attr_b_value = (bool) value;
		attr.set_value(&attr_b_value);		
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveTrigBunchNum
// 
// description : 	Write WaveTrigBunchNum attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveTrigBunchNum(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveTrigBunchNum(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveTrigBunchNum(): " << attr_WaveTrigBunchNum_write << endl;
	
	if (realtimeControl) {
		attr.get_write_value(attr_WaveTrigBunchNum_write);	
		char value;
		if (attr_WaveTrigBunchNum_write == true)
			value = 1;
		else
			value = 0;
		a2605_set_wave_trig_bnc(psIndex, (long)attr_WaveTrigBunchNum_write, a2605);		
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveTrigIrq
// 
// description : 	Extract real attribute values for WaveTrigIrq acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveTrigIrq(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveTrigIrq(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		char value = a2605_get_wave_trig_irq(psIndex, a2605);
		attr_b_value = (bool) value;
		attr.set_value(&attr_b_value);	
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveTrigIrq
// 
// description : 	Write WaveTrigIrq attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveTrigIrq(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveTrigIrq(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveTrigIrq(): " << attr_WaveTrigIrq_write << endl;

	if (realtimeControl) {
		attr.get_write_value(attr_WaveTrigIrq_write);	
		char value;
		if (attr_WaveTrigIrq_write == true)
			value = 1;
		else
			value = 0;
		a2605_set_wave_trig_irq(psIndex, value, a2605);	
	}
	
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_WaveTrigSw
// 
// description : 	Extract real attribute values for WaveTrigSw acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_WaveTrigSw(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_WaveTrigSw(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		char value = a2605_get_wave_trig_sw(psIndex, a2605);
		attr_b_value = (bool) value;
		attr.set_value(&attr_b_value);
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_WaveTrigSw
// 
// description : 	Write WaveTrigSw attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_WaveTrigSw(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_WaveTrigSw(Tango::WAttribute &attr) entering... "<< endl;
	
	INFO_STREAM << "A2605::write_WaveTrigSw(): " << attr_WaveTrigSw_write << endl;

	if (realtimeControl) {
		attr.get_write_value(attr_WaveTrigSw_write);	
		char value;
		if (attr_WaveTrigSw_write == true)
			value = 1;
		else
			value = 0;

		a2605_set_wave_trig_sw(psIndex, value, a2605);
	}
	
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_CurrentControl
// 
// description : 	Extract real attribute values for CurrentControl acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_CurrentControl(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_CurrentControl(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		attr_s_value = a2605_get_current_control(psIndex, a2605);	
	}
	else {
		attr_s_value = A2605_CURR_LOCAL_CTRL  - 1;
	}
	attr.set_value(&attr_s_value);

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_CurrentControl
// 
// description : 	Write CurrentControl attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_CurrentControl(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_CurrentControl(Tango::WAttribute &attr) entering... "<< endl;

	INFO_STREAM << "Psc::write_CurrentControl(): " << attr_CurrentControl_write << endl;
	
	if (realtimeControl) {
		attr.get_write_value(attr_CurrentControl_write);
		if (a2605_set_current_control(psIndex, (long)attr_CurrentControl_write, a2605) < 0) {
			Tango::Except::throw_exception (
				(const char *)"Current control mode not recognized",
				(const char *)"Setting current control failed",
				(const char *)"A2605::write_CurrentControl()");		
		}
	}		
	
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_Waveform
// 
// description : 	Extract real attribute values for Waveform acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_Waveform(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_Waveform(Tango::Attribute &attr) entering... "<< endl;
	
	if (realtimeControl) {
		attr.set_value(&dwave_ps[0], wave_ps_size);	
	}

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_Waveform
// 
// description : 	Write Waveform attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_Waveform(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_Waveform(Tango::WAttribute &attr) entering... "<< endl;

	if (realtimeControl) {	
		const double *waveform;
		long size;	
	
		INFO_STREAM << "A2605::write_Waveform()" << endl;	
	
		attr.get_write_value(waveform);
		size = attr.get_write_value_length();
	
		for (int i = 0; i < size; i++) {
			wave_ps[i] = (long)(waveform[i]*1000000);
		}

		if (a2605_set_wave(psIndex, wave_ps, size, a2605) < 0) {
			Tango::Except::throw_exception (
				(const char *)"Failed to load waveform",
				(const char *)"Waveform not loaded",
				(const char *)"A2605::write_Waveform()");	
		}

		for (int i = 0; i < size; i++) {
			dwave_ps[i] = waveform[i];
		}
		wave_ps_size = size;

	}

}


//+----------------------------------------------------------------------------
//
// method : 		A2605::read_Current
// 
// description : 	Extract real attribute values for Current acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_Current(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_Current(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		long data;
		unsigned long bunchnumber;
		struct timeval buf_time;
		struct timeval tv_now;
		
		gettimeofday(&tv_now,NULL);
	
		check_comm();
		read_last(&data, &buf_time, &bunchnumber, buf[A2605_CURRREAD_IDX]);
		/* force set attribute */
		double set_curr = ((double)(a2605->ps[psIndex].iset_rd))/1000000;
		if (attr_set_current)
			attr_set_current->set_write_value(set_curr);
	
		*attr_Current_read = ((double)data)/1000000;

		/* above threshold */
		if ((diff_err == false) && (fabs(*attr_Current_read - set_curr) > diffCurrErr)) {
			curr_read_diff_time.tv_sec = tv_now.tv_sec;
			curr_read_diff_time.tv_usec = tv_now.tv_usec;
			diff_err = true;
		}
		/* normal condition */
		if (fabs(*attr_Current_read - set_curr) <= diffCurrErr)	{	
			curr_read_diff_time.tv_sec = tv_now.tv_sec;
			curr_read_diff_time.tv_usec = tv_now.tv_usec;			
			diff_err = false;		
		}		
		
	}
	else {
		get_udp_current(attr_Current_read);
		if (attr_set_current)
			attr_set_current->set_write_value(attr_Current_write);
	}



	attr.set_value(attr_Current_read);

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::write_Current
// 
// description : 	Write Current attribute values to hardware.
//
//-----------------------------------------------------------------------------
void A2605::write_Current(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "A2605::write_Current(Tango::WAttribute &attr) entering... "<< endl;

	double tmp;
	attr.get_write_value(tmp);	

	if (attr_Current_write == NAN) {
		Tango::Except::throw_exception (
			(const char *)"Failed to set current",
			(const char *)"Not allowed value (NAN)",
			(const char *)"A2605::write_Current()");	
	}

	if ((get_state() != (Tango::ON)) && (get_state() != (Tango::RUNNING)))  {
		Tango::Except::throw_exception (
			(const char *)"Failed to set current",
			(const char *)"Power supply not in ON/MOVING state",
			(const char *)"A2605::write_Current()");
	}

	if (fabs(tmp) > psModel) {
		Tango::Except::throw_exception (
			(const char *)"Current value not allowed",
			(const char *)"Failed to retreive last data",
			(const char *)"A2605::write_Current()");
	}

	if (realtimeControl) {
		if (a2605_set_current(psIndex, (long)(tmp*1000000), a2605) < 0) {
			Tango::Except::throw_exception (
				(const char *)"Invalid power supply status",
				(const char *)"Setting current failed",
				(const char *)"A2605::write_Current()");	
		}
	}
	else {
		set_udp_current(tmp);
	}

	attr_Current_write = tmp;

	
}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_Voltage
// 
// description : 	Extract real attribute values for Voltage acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_Voltage(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_Voltage(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		long data;
		unsigned long bunchnumber;
		struct timeval buf_time;
	
		check_comm();
		read_last(&data, &buf_time, &bunchnumber, buf[A2605_VOLTAGE_IDX]);
		*attr_Voltage_read = ((double)data)/1000000;
	}
	else {
		get_udp_voltage(attr_Voltage_read);
	}	

	attr.set_value(attr_Voltage_read);	

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_VdcLink
// 
// description : 	Extract real attribute values for VdcLink acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_VdcLink(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_VdcLink(Tango::Attribute &attr) entering... "<< endl;

	if (realtimeControl) {
		long data;
		unsigned long bunchnumber;
		struct timeval buf_time;

		check_comm();
		read_last(&data, &buf_time, &bunchnumber, buf[A2605_VDCLINK_IDX]);
		*attr_VdcLink_read = data;

	}
	else {
		get_udp_voltage(attr_VdcLink_read);
	}

	attr.set_value(attr_VdcLink_read);	

}

//+----------------------------------------------------------------------------
//
// method : 		A2605::read_PsStat
// 
// description : 	Extract real attribute values for PsStat acquisition result.
//
//-----------------------------------------------------------------------------
void A2605::read_PsStat(Tango::Attribute &attr)
{
	DEBUG_STREAM << "A2605::read_PsStat(Tango::Attribute &attr) entering... "<< endl;

	check_comm();
 
	attr.set_value(PsStat, A2605_NUM_STAT, 0, false);
	
}



//+------------------------------------------------------------------
/**
 *	method:	A2605::on
 *
 *	description:	method to execute "On"
 *
 *
 */
//+------------------------------------------------------------------
void A2605::on()
{
	DEBUG_STREAM << "A2605::on(): entering... !" << endl;

	//	Add your own code to control device here
	if (realtimeControl) {
		a2605_set_onoff(psIndex, 1, a2605);	
	}
	else {
		set_udp_onoff(true);
	}

}

//+------------------------------------------------------------------
/**
 *	method:	A2605::off
 *
 *	description:	method to execute "Off"
 *
 *
 */
//+------------------------------------------------------------------
void A2605::off()
{
	DEBUG_STREAM << "A2605::off(): entering... !" << endl;

	//	Add your own code to control device here
	/* abort cycling loop */
	cycle_abort = true;
	
	if (off_run == false) {
		off_abort = false;
		offloop = new offthread::offthread(this);
		offloop->start();
	}

}

//+------------------------------------------------------------------
/**
 *	method:	A2605::reset
 *
 *	description:	method to execute "Reset"
 *
 *
 */
//+------------------------------------------------------------------
void A2605::reset()
{
	DEBUG_STREAM << "A2605::reset(): entering... !" << endl;

	//	Add your own code to control device here
	if (realtimeControl) {
		long stat = a2605_get_status(psIndex, a2605);
	
		if (stat & A2605_RD_STAT_COMM_TOUT)
			reset_dsp();
		
		cycle_abort = true;
		off_abort = true;
		
		a2605_reset(psIndex, a2605);
	}
	else {
		set_udp_reset();
	}

}

//+------------------------------------------------------------------
/**
 *	method:	A2605::get_current
 *
 *	description:	method to execute "GetCurrent"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *A2605::get_current(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();
	DEBUG_STREAM << "A2605::get_current(): entering... !" << endl;

	//	Add your own code to control device here

	if (realtimeControl) {
		long buf_idx[2], num_samples;
		unsigned char unique;
	
		check_comm();
		
		check_get_data_param(argin, buf[A2605_CURRREAD_IDX], buf_idx, &num_samples,&unique);

		argout->length(num_samples);

		/* get last value */
		read_data_calib(unique, buf_idx, &(*argout)[0], buf[A2605_CURRREAD_IDX],0,0.000001,0);

	}
	else {
		argout->length(1);
	}

	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	A2605::get_voltage
 *
 *	description:	method to execute "GetVoltage"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *A2605::get_voltage(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();
	DEBUG_STREAM << "A2605::get_voltage(): entering... !" << endl;

	//	Add your own code to control device here
	if (realtimeControl) {
		long buf_idx[2], num_samples;
		unsigned char unique;

		check_comm();
	
		check_get_data_param(argin, buf[A2605_VOLTAGE_IDX], buf_idx, &num_samples,&unique);

		argout->length(num_samples);

		/* get last value */
		read_data_calib(unique, buf_idx, &(*argout)[0], buf[A2605_VOLTAGE_IDX],0,0.000001,0);

	}
	else {
		argout->length(1);
	}

	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	A2605::get_vdc_link
 *
 *	description:	method to execute "GetVdcLink"
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *A2605::get_vdc_link(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();
	DEBUG_STREAM << "A2605::get_vdc_link(): entering... !" << endl;

	//	Add your own code to control device here
	if (realtimeControl) {
		long buf_idx[2], num_samples;
		unsigned char unique;
	
		check_comm();
	
		check_get_data_param(argin, buf[A2605_VDCLINK_IDX], buf_idx, &num_samples,&unique);

		argout->length(num_samples);

		/* get last value */
		read_data_calib(unique, buf_idx, &(*argout)[0], buf[A2605_VDCLINK_IDX],0,1,0);



	}
	else {
		argout->length(1);
	}

	return argout;
	
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::check_comm
 *
 *	description:	check communication
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
void A2605::check_comm(void)
{

	if (!(comm_ok)) {
		Tango::Except::throw_exception (
			(const char *)"Communication timeout",
			(const char *)"Failed to communicate to device",
			(const char *)"A2605::check_comm()");		
	}

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::set_DevStat
 *
 *	description:	set device status
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
void A2605::set_DevStat(void)
{
	long stat;

	struct timeval now;

	gettimeofday(&now,NULL);
	if (abs((now.tv_sec-last_data_time.tv_sec)*1000000 + (now.tv_usec-last_data_time.tv_usec)) < A2605_STAT_TOUT)
		return;

	last_data_time.tv_sec = now.tv_sec;
	last_data_time.tv_usec = now.tv_usec;

	if (realtimeControl) {
		stat = a2605_get_status(psIndex, a2605);
	}
	else {
		try {
			get_udp_stat(&stat);
			last_sock_time.tv_sec = now.tv_sec;
			last_sock_time.tv_usec = now.tv_usec;
			stat &= ~A2605_RD_STAT_COMM_TOUT;
		}
		catch(Tango::DevFailed e) {
			stat |= A2605_RD_STAT_COMM_TOUT;
				if ((now.tv_sec-last_sock_time.tv_sec) > A2605_STAT_TOUT) {
					delete_udp_socket();
					create_udp_socket();
				}
		}
	}

	std::stringstream s;


	if (stat & A2605_RD_STAT_COMM_TOUT) {
		s << "FAULT: communication timeout"  << endl;
		set_state(Tango::FAULT);
		comm_ok = false;
	}
	else {
		comm_ok = true;
	
		if (A2605_RD_STAT_ONOFF & stat) {
			if (A2605_RD_STAT_WAVE_STAT & stat) {
				set_state(Tango::RUNNING);
				s << "Waveform generation is running" << endl;
			}
			else {
				set_state(Tango::ON);
				s << "Power supply is ON" << endl;
			}
		}
		else {
			set_state(Tango::OFF);
			s << "Power supply is OFF" << endl;	
			if (!realtimeControl) {
				attr_Current_write = 0;
			}
		}
	
		if (!(A2605_RD_STAT_RTMODE & stat)) {
			s << "Realtime communication disabled" << endl;			
		}

		if (realtimeControl) {
			if (A2605_RD_STAT_CYCLE & stat) {
				set_state(Tango::MOVING);
				if (degauss_flag)
					s << "Power supply is DEGAUSSING" << endl;
				else
					s << "Power supply is CYCLING" << endl;	
			}
		}
		else {
			if (udp_cycle_flag) {
				set_state(Tango::MOVING);
				s << "Power supply is CYCLING" << endl;				
			}
		}

		/* reset diff time during loops */
		if (get_state() == (Tango::MOVING)) { 			
			curr_read_diff_time.tv_sec = now.tv_sec;
			curr_read_diff_time.tv_usec = now.tv_usec;	
		}
		

		/* ALARMS */
		if ((A2605_RD_STAT_RTMODE & stat) && (A2605_RD_STAT_BN_ERR & stat)) {
			//set_state(Tango::ALARM);
			s << "Bunch number misalignment" << endl;			
		}	

		/* FAULTS */
		if ((stat & A2605_RD_STAT_FAULT) && (psModel == 20) && (!((stat & 0xff) >> 2))) {
			s << "WARNING: DC bulk not operating correctly" << endl;
		}
		if (stat & A2605_RD_STAT_UNDER_DC) {
			s << "FAULT: dc undervoltage" << endl;
			set_state(Tango::FAULT);
		}
		if (stat & A2605_RD_STAT_MOS_OVT) {
			s << "FAULT: mosfet overtemperature" << endl;
			set_state(Tango::FAULT);
		}
		if (stat & A2605_RD_STAT_SHUNT_OVT) {
			s << "FAULT: shunt overtemperature" << endl;
			set_state(Tango::FAULT);
		}
		if (stat & A2605_RD_STAT_ILOCK) {
			s << "FAULT: interlock occurred" << endl;
			set_state(Tango::FAULT);
		}
		if (stat & A2605_RD_STAT_OVERCUR) {
			s << "FAULT: input overcurrent" << endl;
			set_state(Tango::FAULT);
		}
		if (stat & A2605_RD_STAT_CROWBAR) {
			s << "FAULT: crowbar (short circuit)" << endl;
			set_state(Tango::FAULT);
		}
	} /* else comm ok */

	/* check if not in local control */
	if  (((diff_err == true) && (now.tv_sec - curr_read_diff_time.tv_sec) > diffTimeErr) && 
	     ((A2605_RD_STAT_ONOFF & stat) != 0) && 
			 ((A2605_RD_STAT_WAVE_STAT & stat) == 0) &&
		   ((A2605_CURR_LOCAL_CTRL-1) == a2605_get_current_control(psIndex, a2605))) {
		s << "Read/set current mismatch" << endl;
		set_state(Tango::FAULT);	
	}


	/* array of booleans */
	for (int i = 0; i < A2605_NUM_STAT; i++) {
		if (stat & (1 << i))
			PsStat[i] = true;
		else
			PsStat[i] = false;		
	}

	/* set device status */
	set_status(s.str());	

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::init_buffers
 *
 *	description:	Initialize realtime buffers
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
bool A2605::init_buffers(void)
{
	char bufname[50];


	for (int i = 0; i < A2605_STORE_VALUES; i++) {
		sprintf(bufname,"/dev/buffer%d",(int) (i+A2605_STORE_VALUES*psIndex));
		fd_buffer[i] = open(bufname,O_RDWR);
		if(fd_buffer[i] < 0) {
			cout << "A2605: error opening " << bufname << endl;
			return false;
		}
		else {
			cout << "A2605: opening buffer " << bufname << endl;
		}
			
		buf[i] = 0;
		int ret =ioctl(fd_buffer[i], BUFFER_RT_IOC_READ_VSIZE, &vsize[i]);
		if(ret) {
			cout << "A2605:  error in ioctl! fd_buffer=" << fd_buffer[i] << " ret="<< ret << endl;
			close(fd_buffer[i]);
		}
		
		cout << "A2605:  rbuffer vsize="<< vsize[i] << endl;
		if(vsize[i] <= 0) {
			close(fd_buffer[i]);
		}
	
		buf[i] = 0;
		buf[i] = (struct _buffer *) mmap(0, vsize[i], PROT_READ | PROT_WRITE, MAP_SHARED, fd_buffer[i], 0);
	
		if (buf[i] == NULL) {
			cout << "A2605: error in mmap!!!" << endl;
			close(fd_buffer[i]);
		}
	}

	return true;
	
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::delete_buffers
 *
 *	description:	Delete realtime buffers
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
void A2605::delete_buffers(void)
{

	for (int i = 0; i < A2605_STORE_VALUES; i++) {
		if (buf[i] != NULL) {
			munmap((void *) buf[i], vsize[i]);
			if (fd_buffer[i])
				close(fd_buffer[i]);
		}
	}

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::string_explode
 * Explode a string 
*/
//+------------------------------------------------------------------
void A2605::string_explode(string str, string separator, vector<string>* results)
{
	unsigned int found;
	
	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			results->push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results->push_back(str);
	}

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::axtoi
 *
 *	description:	method to execute "axtoi"
 *	Convert hex string to integer value
 *
 *
 */
//+------------------------------------------------------------------
int A2605::axtoi(char *hexStg) {
  int n = 0;         // position in string
  int m = 0;         // position in digit[] to shift
  int count;         // loop index
  int intValue = 0;  // integer value of hex string
  int digit[5];      // hold values to convert
  while (n < 4) {
     if (hexStg[n]=='\0')
        break;
     if (hexStg[n] > 0x29 && hexStg[n] < 0x40 ) //if 0 to 9
        digit[n] = hexStg[n] & 0x0f;            //convert to int
     else if (hexStg[n] >='a' && hexStg[n] <= 'f') //if a to f
        digit[n] = (hexStg[n] & 0x0f) + 9;      //convert to int
     else if (hexStg[n] >='A' && hexStg[n] <= 'F') //if A to F
        digit[n] = (hexStg[n] & 0x0f) + 9;      //convert to int
     else break;
    n++;
  }
  count = n;
  m = n - 1;
  n = 0;
  while(n < count) {
     // digit[n] is value of hex digit at position n
     // (m << 2) is the number of positions to shift
     // OR the bits into return value
     intValue = intValue | (digit[n] << (m << 2));
     m--;   // adjust the position to set
     n++;   // next digit to process
  }
  return (intValue);
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::configure_network
 *
 *	description:	method to execute "configure_network"
 *
 */
//+------------------------------------------------------------------
void A2605::configure_network(void) 
{
	char *pch;int i=0;
	unsigned char mac_ps[6];
	unsigned char ip_ps[4];
	unsigned char ip_cpu[4];
	char buffer[100];

	
	i = 0;
	memset(buffer,0,100);
	strncpy(buffer,psMacAddress.c_str(),100);
	pch = strtok (buffer,":");
	mac_ps[i++] = axtoi(pch);
  while ((pch != NULL) && (i < 6))
  {
    pch = strtok (NULL, ":");	
		if (pch != NULL)
			mac_ps[i++] = axtoi(pch);
		else
			break;
  }

	i = 0;
	memset(buffer,0,100);
	strncpy(buffer,psIp.c_str(),100);
	pch = strtok (buffer,".");
	ip_ps[i++] = atoi(pch);
  while ((pch != NULL) && (i < 4))
  {
    pch = strtok (NULL, ".");	
		if (pch != NULL)
			ip_ps[i++] = atoi(pch);
		else
			break;
  }

	i = 0;	
	memset(buffer,0,100);
	strncpy(buffer,cpuIp.c_str(),100);
	pch = strtok (buffer,".");
	ip_cpu[i++] = atoi(pch);
  while ((pch != NULL) && (i < 4))
  {
    pch = strtok (NULL, ".");	
		if (pch != NULL)
			ip_cpu[i++] = atoi(pch);
		else
			break;
  }		

	printf("\n ethernet configuration: %s\n",this->get_name().c_str());
	printf(" cpu ip: %d.%d.%d.%d\n",ip_cpu[0],ip_cpu[1],ip_cpu[2],ip_cpu[3]);
	printf(" cpu udp port: %d\n",cpuUdpPort);	
	printf(" ps macaddress: 0x%02x:0x%02x:0x%02x:0x%02x:0x%02x:0x%02x\n",
		mac_ps[0],mac_ps[1],mac_ps[2],mac_ps[3],mac_ps[4],mac_ps[5]);	
	printf(" ps ip: %d.%d.%d.%d\n",ip_ps[0],ip_ps[1],ip_ps[2],ip_ps[3]);
	printf(" ps udp port: %d\n\n",psUdpPort);	
	

	a2605_configure_eth(psIndex, mac_ps, ip_ps, ip_cpu, psUdpPort, cpuUdpPort, a2605);
	a2605_set_ip_checksum(psIndex, a2605);

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::wave_start_sw
 *
 *	description:	method to execute "WaveStartSw"
 *	Start waveform by a software trigger.
 *
 *
 */
//+------------------------------------------------------------------
void A2605::wave_start_sw()
{
	DEBUG_STREAM << "A2605::wave_start_sw(): entering... !" << endl;

	//	Add your own code to control device here
	a2605_start_sw(psIndex, a2605);
	
}

//+------------------------------------------------------------------
/**
 *	method:	A2605::start_cycling
 *
 *	description:	method to execute "StartCycling"
 *	Start cycling procedure
 *
 *
 */
//+------------------------------------------------------------------
void A2605::start_cycling()
{
	DEBUG_STREAM << "A2605::start_cycling(): entering... !" << endl;

	//	Add your own code to control device here

	
	if (get_state() != (Tango::ON)) {
		Tango::Except::throw_exception (
			(const char *)"Failed to start cycling",
			(const char *)"Power supply not in ON state",
			(const char *)"A2605::start_cycling()");
	}

	if (cycle_run == false) {
		cycle_abort = false;
		degauss_flag = false;
		cycleloop = new cyclethread::cyclethread(this);
		cycleloop->start();
	}
		
}

//+------------------------------------------------------------------
/**
 *	method:	A2605::abort
 *
 *	description:	method to execute "Abort"
 *	Abort a cycling or a waveform generation
 *
 *
 */
//+------------------------------------------------------------------
void A2605::abort()
{
	DEBUG_STREAM << "A2605::abort(): entering... !" << endl;

	//	Add your own code to control device here
	cycle_abort = true;
	off_abort = true;
	
	if (realtimeControl) {
		a2605_stop_waveform(psIndex, a2605);
	}

}



//+------------------------------------------------------------------
/**
 *	method:	A2605::get_current_set
 *
 *	description:	method to execute "GetCurrentSet"
 *	Get current set values
 *
 * @param	argin	
 * @return	
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleArray *A2605::get_current_set(const Tango::DevVarLongArray *argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarDoubleArray	*argout  = new Tango::DevVarDoubleArray();

	DEBUG_STREAM << "A2605::get_current_set(): entering... !" << endl;

	if (realtimeControl) {
		//	Add your own code to control device here
		long buf_idx[2], num_samples;
		unsigned char unique;
	
		check_comm();
	
		check_get_data_param(argin, buf[A2605_CURRSET_IDX], buf_idx, &num_samples,&unique);

		argout->length(num_samples);

		/* get last value */
		read_data_calib(unique, buf_idx, &(*argout)[0], buf[A2605_CURRSET_IDX],0 ,0.000001 ,0);
	}
	else {
		argout->length(1);
	}

	return argout;

}




//+------------------------------------------------------------------
/**
 *	method:	A2605::write_cmd
 *
 */
//+------------------------------------------------------------------
long A2605::write_cmd(char *strin, char *strout, int insize) 
{
	int num;

	if ((num = write(sock, strin, insize)) < 0) {
		TangoSys_MemStream out_stream;
		out_stream << strerror(errno) << ends;	
		ERROR_STREAM << out_stream.str() << endl;
		Tango::Except::throw_exception(
			(const char *) "Network communication error",
			out_stream.str(),
			(const char *) "A2605::write_cmd", Tango::ERR);
	}

	usleep(100000);	

	if ((num = recv(sock, strout, A2605_MSG_LEN,0)) < 0) {
		TangoSys_MemStream out_stream;
		ERROR_STREAM << out_stream.str() << endl;			
		Tango::Except::throw_exception(
			(const char *) "Network communication error",
			out_stream.str(),
			(const char *) "A2605::write_cmd", Tango::ERR);
	}

	
	sock_flush();

	return num;

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::sock_connect() 
 *
 */
//+------------------------------------------------------------------
int A2605::sock_connect() 
{
	int connect_err_flag = 0, arg;
	TangoSys_MemStream out_stream;
	

	mysocket.sin_port = htons(A2605_REBOOT_DSP_PORT); 
	mysocket.sin_addr.s_addr = inet_addr(psIp.c_str());
	mysocket.sin_family = AF_INET; 
	if (( sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
		connect_err_flag = 1;
		cout << "Open device " << psIp << " failed, " << strerror(errno) << ends;			
		set_status(out_stream.str());	
		return connect_err_flag;
	}
	else
		connect_err_flag = 0;	
		
	/* getting socket parameters */
	if( (arg = fcntl(sock, F_GETFL, NULL)) < 0) {
		connect_err_flag = 1;
		cout << "Reading socket setting " << psIp << " failed, " << strerror(errno) << ends;		
		set_status(out_stream.str());
		return connect_err_flag;		
	}
	else
		connect_err_flag = 0;
			
	/* setting non blocking socket */						
	arg |= O_NONBLOCK;		
	if (fcntl (sock, F_SETFL, arg) < 0) {			
		connect_err_flag = 1;
		cout << "Setting socket " << psIp << " failed, " << strerror(errno) << ends;		
		set_state(Tango::FAULT); 
		set_status(out_stream.str());
		return connect_err_flag;		
	}
	else
		connect_err_flag = 0;	

	/* connect to device */		
	/* do not generate exception because non blocking socket configured */
	connect ( sock, (const sockaddr*)&mysocket, sizeof(mysocket));
	
	cout << "Connected to " << psIp << endl;		

	sock_flush();

	return connect_err_flag;

}	

//+------------------------------------------------------------------
/**
 *	method:	A2605::sock_flush
 *
 */
//+------------------------------------------------------------------
void A2605::sock_flush() 
{
	char trash[A2605_MSG_LEN];	
	int count=0;
	while (recv(sock, trash, 100,0) >= 0) {
		count++;
		if (count > 10)
			break;
	};	/* flush socket */
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::reset_dsp
 *
 */
//+------------------------------------------------------------------
void A2605::reset_dsp() 
{
	char strin[A2605_MSG_LEN], strout[A2605_MSG_LEN];

	if (sock_connect()) {
		Tango::Except::throw_exception(
			(const char *) "Can't reset dsp",
			(const char *) "Failed to connect to lantronix",
			(const char *) "A2605::reset_dsp", Tango::ERR);
	}

	memset(strin,0,A2605_MSG_LEN);
	memset(strout,0,A2605_MSG_LEN);	
	strin[0]=25;strin[1]=7;strin[5]=7;	
	write_cmd(strin, strout,9);
	usleep(10000);

	memset(strin,0,A2605_MSG_LEN);
	memset(strout,0,A2605_MSG_LEN);	
	strin[0]=26;strin[1]=7;	
	write_cmd(strin, strout,9);
	usleep(10000);

	memset(strin,0,A2605_MSG_LEN);
	memset(strout,0,A2605_MSG_LEN);	
	strin[0]=27;strin[1]=7;	
	write_cmd(strin, strout,9);
	usleep(10000);

	memset(strin,0,A2605_MSG_LEN);
	memset(strout,0,A2605_MSG_LEN);	
	strin[0]=27;strin[1]=7;	
	write_cmd(strin, strout,9);
	usleep(10000);

	if (sock)
		close(sock);

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::create_udp_socket
 */
//+------------------------------------------------------------------
int A2605::create_udp_socket(void)
{
	/*struct used for binding the socket to a local address*/
	int flags, blocking = 1;
	
	/*create the socket*/
	if ((fd_socket=socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
		cout << "Connecting to device" << psIp << " failed, " << strerror(errno) << endl;
		return -1;
	}

	/*init the host_address the socket is beeing bound to*/
	memset((void*)&host_address, 0, sizeof(host_address));
	/*set address family*/
	host_address.sin_family = AF_INET;
	/*accept any incoming messages:*/
	/*host_address.sin_addr.s_addr = INADDR_ANY;*/
	/*the port the socket i to be bound to:*/
	host_address.sin_port=htons(cpuUdpPort);

	/* set block/no block */
  flags = fcntl(fd_socket, F_GETFL, 0);
  if (flags < 0) return -1;
  flags = blocking ? (flags&~O_NONBLOCK) : (flags|O_NONBLOCK);
	if (fcntl(fd_socket, F_SETFL, flags) < 0) {
		cout << "Reading configuration device " << psIp << " failed, " << strerror(errno) << endl;
		return -1;
	}

	/* set socket timeout */
	struct timeval tv;
	 tv.tv_sec = 0;
	 tv.tv_usec = 100000;
	if (setsockopt(fd_socket, SOL_SOCKET, SO_RCVTIMEO, (struct timeval *)&tv, sizeof(struct timeval)) != 0) {
		cout << "Error setting " << psIp <<" timeout, "<< 	strerror(errno) << endl;
		return -1;
 	}

	/* bind it...*/
	if (bind(fd_socket, (struct sockaddr*)&host_address, sizeof(host_address)) < 0) {
		cout << "Binding " << psIp <<" failed, "<< 	strerror(errno) << endl;
		return -1;
	}

	host = (struct hostent *) gethostbyname((char *)psIp.c_str());
	/*init target address structure*/
	target_host_address.sin_family = AF_INET;
	target_host_address.sin_port = htons(psUdpPort);
	
	target_host_address.sin_addr = *((struct in_addr *)host->h_addr);

	return 0;

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::create_udp_socket
 */
//+------------------------------------------------------------------
void A2605::send_udp_socket(char *buffer)
{
	in_mutex->lock();

	/*send it*/
	if ((sendto(fd_socket, buffer, strlen(buffer), 0, 
		(struct sockaddr*)&target_host_address, sizeof(target_host_address))) < 0) {
		string err(strerror(errno));
		INFO_STREAM << "send_udp_socket error:" << err <<endl;
		in_mutex->unlock();
		Tango::Except::throw_exception(
			(const char *) "Send data error",
			err,
			(const char *) "A2605::send_udp_socket", Tango::ERR);
	}
	INFO_STREAM << "send_udp_socket: TX: " << buffer <<endl;	
	
}

//+------------------------------------------------------------------
/**
 *	method:	A2605::receive_udp_socket
 */
//+------------------------------------------------------------------
void A2605::receive_udp_socket(char *buffer, int *bytes_read)
{
	
	if ((*bytes_read = recvfrom(fd_socket,buffer,A2605_MSG_LEN,0,
	                    (struct sockaddr *)&client_addr, &addr_len)) <= 0) {	
		string err(strerror(errno));
		INFO_STREAM << "receive_udp_socket:" << err <<endl;	
		in_mutex->unlock();																			
		Tango::Except::throw_exception(
			(const char *) "Receive data error",
			err,
			(const char *) "A2605::receive_udp_socket", Tango::ERR);
	}
	INFO_STREAM << "receive_udp_socket: RX: " << buffer <<endl;
		
	in_mutex->unlock();
	
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::delete_udp_socket
 */
//+------------------------------------------------------------------
int A2605::delete_udp_socket(void)
{
	if (fd_socket > 0)
		close(fd_socket);
	return 0;
}

//+------------------------------------------------------------------
/**
 *	method:	A2605::check_set_response
 * Check reply command for a setting
 */
//+------------------------------------------------------------------
void A2605::check_set_response(char *cmd_resp)
{

	if ((strlen(cmd_resp) < 4) || (memcmp(cmd_resp,"#AK\r",4) != 0)) {
		Tango::Except::throw_exception(
			(const char *) "Check command response failed",
			(const char *) "Command failed",
			(const char *) "A2605::check_set_response", Tango::ERR);
	}


}

//+------------------------------------------------------------------
/**
 *	method:	A2605::check_get_response
* Check reply command for a reading
 */
//+------------------------------------------------------------------
void A2605::check_get_response(char *get_cmd, char *get_resp)
{

	 /* get cmd: MRI\r, resp:#MRI:1.2\r */

	if ((strlen(get_resp) < 4) || (strlen(get_resp) < 4) || (memcmp(get_cmd,get_resp+1,3) != 0)) {
		Tango::Except::throw_exception(
			(const char *) "Check reading response failed",
			(const char *) "Reading failed",
			(const char *) "A2605::check_get_response", Tango::ERR);
	}

}



//+------------------------------------------------------------------
/**
 *	method:	A2605::get_udp_current
 * Get current value
 */
//+------------------------------------------------------------------
void A2605::get_udp_current(Tango::DevDouble *current)
{
	char tx_buffer[A2605_MSG_LEN],rx_buffer[A2605_MSG_LEN];
	int bytes_read;

	sprintf(tx_buffer,"MRI\r\0");
	send_udp_socket(tx_buffer);

	memset(rx_buffer,0,A2605_MSG_LEN);
	receive_udp_socket(rx_buffer, &bytes_read);

	/* response: #MRI:+0.1234\r */
	if (bytes_read < 7) {
		Tango::Except::throw_exception(
			(const char *) "Failed to read status",
			(const char *) "Data corrupted",
			(const char *) "A2605::get_udp_stat", Tango::ERR);
	}

	sscanf(rx_buffer+5,"%lf\r",current);


}


//+------------------------------------------------------------------
/**
 *	method:	A2605::get_udp_voltage
 * Get voltage
 */
//+------------------------------------------------------------------
void A2605::get_udp_voltage(Tango::DevDouble *voltage)
{

	char tx_buffer[A2605_MSG_LEN],rx_buffer[A2605_MSG_LEN];
	int bytes_read;

	sprintf(tx_buffer,"MRV\r\0");
	send_udp_socket(tx_buffer);
	memset(rx_buffer,0,A2605_MSG_LEN);
	receive_udp_socket(rx_buffer, &bytes_read);

	/* response: #MRI:+0.1234\r */
	if (bytes_read < 7) {
		Tango::Except::throw_exception(
			(const char *) "Failed to read status",
			(const char *) "Data corrupted",
			(const char *) "A2605::get_udp_stat", Tango::ERR);
	}

	sscanf(rx_buffer+5,"%lf\r",voltage);

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::get_udp_dclink
 * Get dclink voltage
 */
//+------------------------------------------------------------------
void A2605::get_udp_dclink(Tango::DevDouble *dclink)
{

	char tx_buffer[A2605_MSG_LEN],rx_buffer[A2605_MSG_LEN];
	int bytes_read;

	sprintf(tx_buffer,"MRP\r\0");
	send_udp_socket(tx_buffer);
	memset(rx_buffer,0,A2605_MSG_LEN);
	receive_udp_socket(rx_buffer, &bytes_read);

	/* response: #MRI:+0.1234\r */
	if (bytes_read < 7) {
		Tango::Except::throw_exception(
			(const char *) "Failed to read status",
			(const char *) "Data corrupted",
			(const char *) "A2605::get_udp_stat", Tango::ERR);
	}

	sscanf(rx_buffer+5,"%lf\r",dclink);

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::get_udp_dclink
 * Get dclink voltage
 */
//+------------------------------------------------------------------
void A2605::get_udp_stat(long *stat)
{
	char tx_buffer[A2605_MSG_LEN],rx_buffer[A2605_MSG_LEN];
	int bytes_read;

	sprintf(tx_buffer,"MST\r\0");
	send_udp_socket(tx_buffer);
	memset(rx_buffer,0,A2605_MSG_LEN);
	receive_udp_socket(rx_buffer, &bytes_read);

	if (bytes_read != 8) {
		Tango::Except::throw_exception(
			(const char *) "Failed to read status",
			(const char *) "Data corrupted",
			(const char *) "A2605::get_udp_stat", Tango::ERR);
	}
	check_get_response(tx_buffer,rx_buffer);

	/* response: #MST:FE\r */

	if ((char_to_hex[rx_buffer[5]]== 0) || (char_to_hex[rx_buffer[6]] == 0)) {
		Tango::Except::throw_exception(
			(const char *) "Failed to read status",
			(const char *) "Status corrupted",
			(const char *) "A2605::get_udp_stat", Tango::ERR);
	}

	*stat = ((char_to_hex[rx_buffer[5]]-1) << 4) + (char_to_hex[rx_buffer[6]]-1);
	
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::set_udp_current
 * Set current
 */
//+------------------------------------------------------------------
void A2605::set_udp_current(Tango::DevDouble current)
{

	char tx_buffer[A2605_MSG_LEN],rx_buffer[A2605_MSG_LEN];
	int bytes_read;

	if (slewRateLimit) {
		sprintf(tx_buffer,"MRM:%.4f\r\0", current);
	}
	else {
		sprintf(tx_buffer,"MWI:%.4f\r\0", current);
	}
	send_udp_socket(tx_buffer);
	memset(rx_buffer,0,A2605_MSG_LEN);
	receive_udp_socket(rx_buffer, &bytes_read);
	check_set_response(rx_buffer);

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::set_udp_onoff
 * Set ps on/off
 */
//+------------------------------------------------------------------
void A2605::set_udp_onoff(bool on_flag)
{

	char tx_buffer[A2605_MSG_LEN],rx_buffer[A2605_MSG_LEN];
	int bytes_read;

	if (on_flag)
		sprintf(tx_buffer,"MON\r\0");
	else
		sprintf(tx_buffer,"MOFF\r\0");
	
	send_udp_socket(tx_buffer);
	memset(rx_buffer,0,A2605_MSG_LEN);
	receive_udp_socket(rx_buffer, &bytes_read);
	check_set_response(rx_buffer);

	if (on_flag == 0) {
		double curr = 0;
		attr_set_current->set_write_value(curr);
		attr_Current_write = curr;
	}

}


//+------------------------------------------------------------------
/**
 *	method:	A2605::set_udp_reset
 * Set ps on/off
 */
//+------------------------------------------------------------------
void A2605::set_udp_reset(void)
{

	char tx_buffer[A2605_MSG_LEN],rx_buffer[A2605_MSG_LEN];
	int bytes_read;

	sprintf(tx_buffer,"MRESET\r\0");
	send_udp_socket(tx_buffer);
	memset(rx_buffer,0,A2605_MSG_LEN);
	receive_udp_socket(rx_buffer, &bytes_read);
	check_set_response(rx_buffer);

	/* switch ps off */
	set_udp_onoff(false);

}




//+------------------------------------------------------------------
/**
 *	method:	A2605::config_polling_attr
 * Set ps on/off
 */
//+------------------------------------------------------------------
void A2605::config_polling_attr(void)
{
	try {

	if ((polling_config == false) && (!realtimeControl)) {
		A2605Class	*ds_class =
			(static_cast<A2605Class *>(get_device_class()));
		Tango::MultiClassAttribute* multiClassAttribute = ds_class->get_class_attr();
		string current_str("Current");
		multiClassAttribute->get_attr(current_str).set_polling_period(1000);
		string voltage_str("Voltage");
		multiClassAttribute->get_attr(voltage_str).set_polling_period(1000);	
		string vdclink_str("VdcLink");		
		multiClassAttribute->get_attr(vdclink_str).set_polling_period(1000);
		polling_config = true;
	}
	
	} catch(...) {} 
	
}






//+------------------------------------------------------------------
/**
 *	method:	A2605::save_config
 *
 *	description:	method to execute "SaveConfig"
 *	Save ps mac address into property
 *
 *
 */
//+------------------------------------------------------------------
void A2605::save_config()
{
	DEBUG_STREAM << "A2605::save_config(): entering... !" << endl;

	//	Add your own code to control device here
	
	Tango::Database *db = new Tango::Database(host_rw,port_rw); 

	/* Ps mac Address */
	try {
		string dev_property("PsMacAddress");
		Tango::DbDatum config(dev_property);
		Tango::DbData db_data_del, db_data_put;
		db_data_del.push_back(config);
		/* start data */
		config << psMacAddress;
		/* end data */
		db_data_put.push_back(config);

		db->put_device_property(get_name(), db_data_put);

	}
	catch (CORBA::Exception &e) {
		delete db;	
		ERROR_STREAM << "A2605::save_config(): Ps mac Address" << endl;		
		Tango::Except::throw_exception (
			(const char *)"Failed to update configuration",
			(const char *)"Can't write into database",
			(const char *)"A2605::save_config");
	}
	
	delete db;	
	
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::start_degaussing
 *
 *	description:	method to execute "StartDegaussing"
 *	Start degaussing procedure
 *
 *
 */
//+------------------------------------------------------------------
void A2605::start_degaussing()
{
	DEBUG_STREAM << "A2605::start_degaussing(): entering... !" << endl;

	//	Add your own code to control device here
	if (get_state() != (Tango::ON)) {
		Tango::Except::throw_exception (
			(const char *)"Failed to start degaussing",
			(const char *)"Power supply not in ON state",
			(const char *)"A2605::start_degaussing()");
	}

	if (cycle_run == false) {
		cycle_abort = false;
		degauss_flag = true;
		cycleloop = new cyclethread::cyclethread(this);
		cycleloop->start();
	}
}


//+------------------------------------------------------------------
/**
 *	method:	A2605::get_cmd_log
 *
 *	description:	method to execute "GetCmdLog"
 *
 * @param	argin	number of last commands
 * @return	report
 *
 */
//+------------------------------------------------------------------
Tango::DevVarStringArray *A2605::get_cmd_log(Tango::DevLong argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	Tango::DevVarStringArray	*argout  = new Tango::DevVarStringArray();

	DEBUG_STREAM << "A2605::get_cmd_log(): entering... !" << endl;

	//	Add your own code to control device here
	if (realtimeControl == false) {
		Tango::Except::throw_exception (
			(const char *)"Failed to get commad log",
			(const char *)"Realtime mode disabled",
			(const char *)"A2605::get_cmd_log()");			
	}


	if (argin > A2605_TX_LOG_SIZE_BUF) {
		Tango::Except::throw_exception (
			(const char *)"Failed to get commad log",
			(const char *)"Requested too much data",
			(const char *)"A2605::get_cmd_log()");			
	}

	int last_index = a2605->ps[psIndex].tx_log_idx-1;
	char log_string[100];	
	
	
	if (last_index < 0)
		last_index = A2605_TX_LOG_SIZE_BUF-1;
		
	int start1,end1,start2,end2;

	argout->length(argin);
	int count = 0;
	
	if (argin <= last_index) {
		start1 = last_index - argin + 1; end1 = last_index;
		start2 = 0; end2 = 0;
	}
	else {
		start2 = 0; end2 = last_index;
		end1 = A2605_TX_LOG_SIZE_BUF - 1;
		start1 = A2605_TX_LOG_SIZE_BUF - argin + last_index + 1;
	}

	for (int i = end1; i >= start1; i--) {
		memset(log_string, 0, 100);
		sprintf(log_string, "bn:%d - bndiff:%d - cmd:",
			(int)a2605->ps[psIndex].log_buf[i].bunch_number,
			(int)(a2605->ps[psIndex].bunch_number - a2605->ps[psIndex].log_buf[i].bunch_number));
			
		if (a2605->ps[psIndex].log_buf[i].status & A2605_WR_STAT_ONOFF)
			sprintf(log_string+strlen(log_string),"ON");
		else	
			sprintf(log_string+strlen(log_string),"OFF");
		if (a2605->ps[psIndex].log_buf[i].status & A2605_WR_STAT_RESET)
			sprintf(log_string+strlen(log_string),"RESET");	
		
		sprintf(log_string+strlen(log_string)," - ctrl:");
			
		if (a2605->ps[psIndex].log_buf[i].rnm_cur_flag)
			sprintf(log_string+strlen(log_string),"REM");		
		else
			sprintf(log_string+strlen(log_string),"LOC");
			
		sprintf(log_string+strlen(log_string)," - cur:%.4f A",(double)a2605->ps[psIndex].log_buf[i].curset/1000000);				 	 

		(*argout)[count] = CORBA::string_dup(log_string);
		
		count++;

	}
	
	if (count < argin) { 
	
		for (int i = end2; i >= start2; i--) {
			memset(log_string, 0, 100);
			sprintf(log_string, "bn:%d - bndiff:%d - cmd:",
				(int)a2605->ps[psIndex].log_buf[i].bunch_number,
				(int)(a2605->ps[psIndex].bunch_number - a2605->ps[psIndex].log_buf[i].bunch_number));
			
			if (a2605->ps[psIndex].log_buf[i].status & A2605_WR_STAT_ONOFF)
				sprintf(log_string+strlen(log_string),"ON");
			else	
				sprintf(log_string+strlen(log_string),"OFF");
			if (a2605->ps[psIndex].log_buf[i].status & A2605_WR_STAT_RESET)
				sprintf(log_string+strlen(log_string),"RESET");	
		
			sprintf(log_string+strlen(log_string)," - ctrl:");
			
			if (a2605->ps[psIndex].log_buf[i].rnm_cur_flag)
				sprintf(log_string+strlen(log_string),"REM");		
			else
				sprintf(log_string+strlen(log_string),"LOC");
			
			sprintf(log_string+strlen(log_string)," - cur:%.4f A",(double)a2605->ps[psIndex].log_buf[i].curset/1000000);				 	 

			(*argout)[count] = CORBA::string_dup(log_string);		
			count++;
		}	
	}

	return argout;

}


}	//	namespace
